
//#include <stdio.h>
#include <iostream>
#include <fstream>

#include "cg.hh"
#include "vcg.hh"
#include "opts.hh"
#include "eopt.hh"
#include "ls.hh"
#include "verbose.hh"
#include "ilpd.hh"

// Global structure for generating DOT graphs
//struct node_writer {
//  void operator()(std::ostream& out, vertex_id v_id) const {
//  	lcc::Node u = get_lcc_node(v_id, g_current_dag);
//    out << "[label=\"" << lcc::opname(u->op) << "\"]";
//  }
//};


bool CG::b_initialized = false;
char *CG::archi = NULL;
Emitter *CG::emitter = NULL;
Architecture *CG::architecture = NULL;
CG *gp_cg = NULL;
TimeProfile *CG::empty_time_profile = NULL;
Flags CG::flags;

#ifdef __cplusplus
extern "C" {
#endif

  void cg_initialize(int argc, char *argv[]) { CG::initialize(argc, argv); }
  void cg_finalize(void) { CG::finalize(); }
  void cg_function(lcc::Symbol f, lcc::Symbol caller[], lcc::Symbol callee[], int ncalls) {
    gp_cg->function(f, caller, callee, ncalls);
  }
  lcc::Node cg_gen(lcc::Node p) { //, Indeg indeg) { 
    assert(gp_cg);
    return gp_cg->gen(p); //, indeg); 
  }
  void cg_emit(lcc::Node p, char *archi) { // , Indeg indeg, char *archi) {
    assert(gp_cg);
    gp_cg->emit(p, archi); //indeg, archi);
  }

#ifdef __cplusplus
}
#endif

CG::CG()
{
  //pu = NULL;
}

extern int ARM_ASSEMBLER_MODE; /* optcg/src/opt.c */
extern int OPT_C_DEBUG; /* optcg/src/opt.c */

void CG::initialize(int argc, char *argv[])
{
  gp_cg = new CG();
  int i_param;
  
  for(i_param = 1; i_param < argc; i_param++) {

    if(argv[i_param][0] != '-') {
      break;
    }

    if(!strcmp(argv[i_param], "-?")) {
      flags.usage();
      exit(0);
    }
    
    if(!strcmp(argv[i_param], "-emitter:mnemonic")){
      Emitter::always_mnemonic = true;
    }
    if(!strcmp(argv[i_param], "-emitter:debug")){
      Emitter::debug_mode = true;
    }
    if(!strcmp(argv[i_param], "-arm_asm"))
    {
      ARM_ASSEMBLER_MODE = 1;
    }
    if(!strcmp(argv[i_param], "-opt_c_debug")){
      OPT_C_DEBUG = 1;
    }
    
    if(!strncmp(argv[i_param], "-archi=", 7)) {
      archi = &argv[i_param][7];
      //Check if the file exists
      FILE *f = fopen(archi, "r");
      if(!f) {
	std::cerr << "\nFATAL ERROR\n  File: " 
		  << archi << " does not exist or is unreadable.\n\n";
	exit(-1);
      }
      fclose(f);
    }

    if(!strncmp(argv[i_param], "-vcg", 4)) {
      flags.b_vcg = true;
    }
    if(!strncmp(argv[i_param], "-ilp", 4)) {
      flags.b_ilp = true;
    }

    if(!strncmp(argv[i_param], "-ddep", 5)) {
      flags.no_ddep_link = true;
    }

    if(!strncmp(argv[i_param], "-dot", 4)) {
      flags.b_dot = true;
    }

    if(!strncmp(argv[i_param], "-verb=", 6)) {
      int verb_level = atoi((char *)&argv[i_param][6]);
      Verbose::verbosityLevel = atoi((char *)&argv[i_param][6]);
    }

    if(!strncmp(argv[i_param], "-opt=", 5)) {
      if(!strncasecmp(&(argv[i_param][5]), "e", 1)) {
	flags.opt_type = Flags::ENERGY;
      }
      if(!strncasecmp(&(argv[i_param][5]), "t", 1)) {
	flags.opt_type = Flags::TIME;
      }
    }

    if(!strncmp(argv[i_param], "-bb=", 4)) {
      flags.i_bb = atoi((char *)&argv[i_param][4]);
    }

    if(!strncmp(argv[i_param], "-hs=", 4)) {
      flags.i_hs = atoi((char *)&argv[i_param][4]);
    }

    if(!strncmp(argv[i_param], "-ts=", 4)) {
      flags.i_ts = atoi((char *)&argv[i_param][4]);
    }

  }

  if(!archi) {
    std::cerr << "\nFATAL ERROR:\n"
      "  Target architecture specification in ADML is required.\n"
      "  Option to lcc: -Wf-archi=<XML file>\n\n";
    exit(-2);
  }
  architecture = new Architecture(archi);
  assert(architecture);
  architecture->parse();
  emitter = new Emitter(architecture);
//   if(architecture) {
//     if(architecture->getSawErrors()) {
//       exit(-1);
//     }
//   }
//  spec->parse_funtional_units();

  //Build empty TimeProfile
  //Get number of fu
  
  //FIXME TODO
  //  int const n = 2;
  //  TimeProfile *init_tp = new TimeProfile(n);

/*  
  empty_time_profile = new TimeProfile();
  empty_time_profile->add_functional_unit("adder1", 5, 2);
  empty_time_profile->add_functional_unit("adder2", 3, 1);
  std::cerr << "Time Profile:\n" << (*empty_time_profile) << std::endl;
  TimeProfile *tp1 = new TimeProfile(*empty_time_profile);
  std::cerr << "Time Profile:\n" << (*tp1) << std::endl;
  delete tp1;
*/
//  solution_space = new SolutionSpace();
  b_initialized = true;
  
}

void CG::finalize(void)
{
  delete gp_cg;
  delete empty_time_profile;
  delete architecture;
  delete emitter;
  //delete solution_space;
  b_initialized = false;
}

/*
//p : parent, q : child
void CG::build_dag(lcc::Node p, DAG & dag)
{
  vertex_t vertex;
  property_map< DAG, vertex_index_t >::type index_map;
  property_map< DAG, vertex_t >::type node_map;
  graph_traits< DAG >::vertex_iterator vi, vi_end;
  vertex_id kid_id;
  lcc::Node v;

  if(p && !p->x.listed) {
    vertex_id p_id = add_vertex(dag);
    put(vertex, dag, p_id, p);
    node_id_map[p] = p_id;
    p->x.listed = 1;

    for(int i = 0; i < 2; i++) {
      build_dag(p->kids[i], dag);
      index_map = get(vertex_index, dag);
      node_map = get(vertex, dag);
      if(p->kids[i]) {
	//	for(tie(vi, vi_end) = vertices(dag); vi != vi_end; ++vi) {
	//	  v = node_map[*vi];
	//	  if(v == p->kids[i]) {
	node_id_map_t::const_iterator v_iter = node_id_map.find(p->kids[i]);
	assert(v_iter != node_id_map.end());
	kid_id = v_iter->second;
	//    kid_id = index_map[*vi];
	add_edge(kid_id, p_id, dag);
      }
    }
  }
    
  //  vertex_id v_id = boost::add_vertex(dag);
  //  vertex_t vertex;
  //  boost::put(vertex, dag, v_id, q);
  //  boost::get(vertex, dag, p)
    
}
*/
lcc::Node CG::gen(lcc::Node p) //, Indeg indeg) {
{
  VERB3 << "CG::gen (DAG) - " << lcc::opname(p->op) << "\n";
  //  DAG dag;
  //node_id_map.clear();
  //dag.clear();
  //Build forest of DAGs
  PU pu = PU(p, architecture, flags);

  //debug
  if(flags.b_vcg || flags.b_dot) {
    pu.print(flags.b_vcg, flags.b_dot);
  }

  //Generate ILP
  if(flags.b_ilp) {

    char file_name[12];
    static int dag_number;
    std::vector< DAG * > dags = pu.get_dags();
    std::vector< DAG * >::iterator d_iter = dags.begin();
    for(; d_iter != dags.end(); ++d_iter) {
      DAG *dag = *d_iter;

      sprintf(file_name, "%s%.2d%s", "ilp", dag_number, ".dat");
      std::ofstream dat_file(file_name);
      dat_file << "data;\n" << std::endl;
      ILPD(*dag, dat_file);
      architecture->ilp(dat_file);
      dat_file.close();

      dag_number++;
    }
  }

  Optimizer *opt;

  switch(flags.opt_type) {
  case Flags::ENERGY: {
    opt = new EOptimist(architecture);
    break;
  }
  case Flags::TIME: {
    opt = new Optimist(architecture);
    break;
  }
  default:
    opt = new Optimist(architecture);
  }

  std::vector< DAG * >::iterator dag_iter = pu.get_dags().begin();
  bool b_single_bb = (flags.i_bb == -1) ? false : true;

  for(; dag_iter != pu.get_dags().end(); ++dag_iter) {
    if(b_single_bb && flags.i_bb != 0) {
      flags.i_bb--;
      continue;
    }
    DAG *bb = *dag_iter;
    Schedule *s = opt->schedule(*bb, flags);
    schedule.push_back(std::make_pair(s, bb));
    if(b_single_bb) {
      flags.i_bb = -2;
    }
  }

  delete opt;  

  /*
  for(lcc::Node q = p; q; q = q->link) {
    build_dag(q, dag);
  }
  for(lcc::Node q = p; q; q = q->link) {
    if(q->link) {
      vertex_id q_id, q_link_id;
      node_id_map_t::const_iterator v_iter = node_id_map.find(q);
      assert(v_iter != node_id_map.end());
      q_id = v_iter->second;
      v_iter = node_id_map.find(q->link);
      assert(v_iter != node_id_map.end());
      q_link_id = v_iter->second;
      add_edge(q_id, q_link_id, dag);
    }
  }

  boost::print_graph(dag);
  boost::write_graphviz(std::cout, dag, node_writer());
  vcg(dag);

  //Add dependences

*/
//    /* Generates forest of DAGs */
//  //   char p_file_name[12];
//  //   sprintf(p_file_name, "%s%d%s", "dag_", u_forest_number, ".vcg");
//  //   std::ofstream f(p_file_name);
//  //   VCG(p, f);
//  //   f.close();
//    /* VCG - end */

//  //   std::cerr << "Before dfs\n";

//    //Build the DAG GRAPH: FIXME -- check for acyclicity
//    //The rank for nodes (start value)

//    static int bb_number = 0;
//    basic_block_id previous_bb = current_bb;
//    for(lcc::Node q = p; q; q = q->link) {
//      switch(generic(q->op)) {
//      case lcc::LABEL:
//        current_bb = pu->add_basic_block(++bb_number); // BB0 entry block
//        std::cerr << "\n### LABEL: " << q->syms[0]->u.l.label << " ###\n";
//        break;
//      default:
//  //        if(previous_bb != current_bb) {
//  //  	pu->add_cfg_edge(previous_bb, current_bb);
//  //  	previous_bb = current_bb;
//  //      }

//      }
//      mark_dfs(q, current_bb);
//    }
//    pu->print();

//    for(lcc::Node q = p; q; q = q->link) {
//      dfs(q);
//    }

//    MARK::iterator v = mark.begin();
//    for(; v != mark.end(); ++v) {
//      //Each node has a uniqu ID (starting from value rank: 0)
//      (g[(*v).first]).id = mark[v->first];
//      if(v->first->kids[0]) {
//        (g[(*v).first]).pred.push_front(v->first->kids[0]);
//        (g[v->first->kids[0]]).succ.push_front(v->first);
//      }
//      if(v->first->kids[1]) {
//        (g[(*v).first]).pred.push_front(v->first->kids[1]);
//        (g[v->first->kids[1]]).succ.push_front(v->first);
//      }
//    }
//    for(lcc::Node q = p; q; q = q->link) {
//      if(q->link) {
//        g[q].succ.push_front(q->link);
//        g[q->link].pred.push_front(q);
//      }
//    }
//    //End build GRAPH

//    //Dump the graph in VCG format if wanted.
//    if(b_vcg)
//      vcg(g);
//    Z z_0;
//    INDEG indeg;
//    v = mark.begin();
//    for(; v != mark.end(); ++v) {
//      std::cerr << "opname: " << lcc::opname(v->first->op) << " rank: " << v->second << endl;
//      lcc::Node w = v->first;
//      assert(w);
//      indeg[w] = 0;
//      if(w->kids[0]) indeg[w]++;
//      if(w->kids[1]) indeg[w]++;
//      //Build z_0
//      if(indeg[w] == 0) {
//        z_0.insert(w);
//      }
//    }
//    for(lcc::Node q = p; q; q = q->link) {
//      if(q->link) indeg[q->link]++;
//    }

//    // Printing
//    INDEG::iterator ii = indeg.begin();
//    for(; ii != indeg.end(); ++ii) {
//      std::cerr << "opname: " << lcc::opname(ii->first->op) << " indeg: " << ii->second
//  	      << " rank: " << mark[ii->first] << endl;
//      lcc::Node w = ii->first;
//      assert(w);
//      indeg[w] = 0;
//      if(w->kids[0]) indeg[w]++;
//      if(w->kids[1]) indeg[w]++;
//    }
//    Z::iterator si = z_0.begin();
//    std::cerr << "z_0:\n";
//    for(; si != z_0.end(); ++si) {
//      std::cerr << lcc::opname((*si)->op) << " " << " rank: " << mark[(*si)] << " ";
//    }
//    std::cerr << endl;
//    //End printing

//    //One step
//    Z::iterator zi = z_0.begin();
//    for(; zi != z_0.end(); ++zi) {
//      Z z_prime;
//      INDEG indeg_prime;
//      select(indeg, z_0, *zi, &indeg_prime, &z_prime);
//    }

//    mark.clear();
//    indeg.clear();
//    z_0.clear();
//    g.clear();

  return p; 
}

/*
void CG::vcg(const GRAPH g)
{
  int static i_bb_number;
  char p_file_name[12];

  sprintf(p_file_name, "%s%.2d%s", "dag", i_bb_number++, ".vcg");
  std::ofstream file(p_file_name);
  VCG(g, file);
  file.close();
}
*/

 /*
void CG::vcg(const DAG g)
{
  int static i_bb_number;
  char p_file_name[12];

  sprintf(p_file_name, "%s%.2d%s", "dag", i_bb_number++, ".vcg");
  std::ofstream file(p_file_name);
  VCG(g, file);
  file.close();
}
 */
/*
void CG::select(INDEG indeg, Z z, lcc::Node v, INDEG *indeg_prime, Z *z_prime)
{
  Z::iterator zi = z.find(v), zi2 = z.begin();
  assert(zi != z.end());

  //Copy z to z_prime and remove v from z_prime
  for(; zi2 != z.end(); ++zi2) {
    if(zi2 != zi)
      z_prime->insert(*zi2);
  }
  //Copy the indegrees
  INDEG::iterator indegi = indeg.begin();
  for(; indegi != indeg.end(); ++indegi) {
    (*indeg_prime)[indegi->first] = indegi->second;
  }
  std::list < lcc::Node >::iterator predi = (g[v]).pred.begin();
  for(; predi != (g[v]).pred.end(); ++predi) {
    if( (--(*indeg_prime)[*predi]) == 0) {
      //*predi enters z_prime
      z_prime->insert(*predi);
    }
  }
}
*/
void CG::function(lcc::Symbol f, 
		  lcc::Symbol *caller,
		  lcc::Symbol *callee,
		  int ncalls)
{
  VERB3 << "#Function\n_" << f->name << "\n";
}

void CG::emit(lcc::Node p, char *archi)
{
  VERB3 << "CG::emit - " << lcc::opname(p->op) << "\n";

  std::list< std::pair< Schedule *, DAG * > >::iterator si;
  for(si = schedule.begin(); si != schedule.end(); ++si) {
    //emit code Anders method class Emitter();
    emitter->emit(si->first);
  }
  schedule.clear();
}

/*
 * void CG::mark_dfs(lcc::Node v, basic_block_id current_bb, node_id _parent)
{
  //  assert(v);
  node_id parent = _parent;
  if(v && v->x.inst == 0) {
    v->x.inst = ++rank;
    node_id child = pu->add_node(rank, v, current_bb);
    if(parent != NOPARENT)
      pu->add_bb_edge(child, parent, current_bb);
    mark_dfs(v->kids[0], current_bb, child);
    mark_dfs(v->kids[1], current_bb, child);
  }
}

void CG::dfs(lcc::Node v)
{
  mark[v] = rank++;
  for(int i = 0; i < 2; i++) {
    if(lcc::Node w = v->kids[i]) {
      MARK::iterator p = mark.find(w);
      if(p == mark.end()) {
	dfs(w);
      }
    }
  }
}

void CG::print_graph(void)
{
  GRAPH::iterator gi = g.begin();
  for(; gi != g.end(); ++gi) {
    std::cout << "Node rank - " << mark[gi->first] << std::endl;
    NODES::iterator i = (g[gi->first]).pred.begin();
    std::cout << "pred list:\n  ";
    for(; i != (g[gi->first]).pred.end(); ++i) {
      std::cout << " " << mark[*i];
    }
    i = (g[gi->first]).succ.begin();
    std::cout << "\nsucc list:\n ";
    for(; i != (g[gi->first]).succ.end(); ++i) {
      std::cout << " " << mark[*i];
    }
    std::cout << std::endl;
  }
}
*/
